@@ -180,6 +180,14 @@ pub enum Error {
   /// Invalid id or number syntax
   #[error("{0}: invalid id syntax: {2:?}: {1}")]
   InvalidIdSyntax(RemoteObjectKind,String,String),
+
+  /// Unsupported remote URL format
+  #[error("unsupported remote URL format: {1:?}: {0}")]
+  UnsupportedRemoteUrlSyntax(String,String),
+
+  /// Remote URL does not match the forge hsotname
+  #[error("remote URL host {url:?} does not match forge host {forge:?}")]
+  UrlHostMismatch { url: String, forge: String },
 }
 
 /// Kind of a thing on a forge.  Used mostly in errors.
@@ -209,7 +217,8 @@ pub enum Kind {
 }
 
 /// Secret access token.
-#[derive(Clone,From,FromStr)]
+#[derive(Clone,From,FromStr,Serialize,Deserialize)]
+#[serde(transparent)]
 pub struct Token(pub String);
 
 impl Debug for Token {
@@ -229,26 +238,42 @@ pub(crate) const FORGES: &[ForgeEntry] = &[
 ];
 
 /// Instructions for how to connect to a forge
-#[derive(Default,Debug)]
+#[derive(Default,Clone,Debug,Serialize,Deserialize)]
 pub struct Config {
-  pub token: Option<Token>,
+  /// Access token (secret).
+  ///
+  /// If left as `None`, `Config::forge()` will do the work
+  /// of `load_default_token`.
+  #[serde(default)]
+  pub token: Option<TokenConfig>,
 
   /// The kind of forge (ie, the protocol to speak).
   ///
   /// Currently, `None` is not supported.  In the future omitting
   /// `kind` might result in auto-guessing from `host`.
+  #[serde(default)]
   pub kind: Option<Kind>,
 
   /// Hostname.
   ///
   /// Eg, `gitlab.com`, `github.com`, `salsa.debian.org`.
+  #[serde(default)]
   pub host: String,
 
   /// Do not specify this field.  Instead, say `..Default::default()`.
   /// New fields may be added and this will not be considered a semver break.
+  #[serde(skip)]
   pub _non_exhaustive: (),
 }
 
+/// Instructions for how to obtain an access token
+#[derive(Clone,Debug,Serialize,Deserialize)]
+pub enum TokenConfig {
+  /** Use anonymous access, do not authenticate. */ Anonymous,
+  /** Read the token from this file.             */ Path(PathBuf),
+  /** Use this token.                            */ Value(Token),
+}
+
 impl Config {
   /// Main constructor for a `Forge`
   #[throws(FE)]
@@ -261,6 +286,22 @@ impl Config {
   }
 }
 
+impl TokenConfig {
+  #[throws(AE)]
+  fn get_token(&self) -> Option<Token> {
+    match self {
+      TokenConfig::Anonymous => None,
+      TokenConfig::Value(v) => Some(v).cloned(),
+      TokenConfig::Path(path) => {
+        let token = Token::load(path)?
+          .ok_or_else(|| anyhow!("specified token file does not exist"))
+          .with_context(|| format!("{:?}", &path))?;
+        Some(token)
+      }
+    }
+  }
+}
+
 impl Token {
   /// Load a secret token from a specified file.
   ///
@@ -334,19 +375,47 @@ impl Config {
     path
   }
 
+  #[throws(AE)]
+  pub(crate) fn get_token_or_default(&self) -> Option<Token> {
+    match &self.token {
+      Some(c) => {
+        c.get_token()?
+      },
+      None => {
+        let path = self.default_token_path()?;
+        let token = Token::load(&path)?;
+        token
+      },
+    }
+
+  }
+
   /// Load the default token, updating this Config.
   ///
-  /// If `self.token` is already `Some()`, does nothing.
-  /// Likewise, if the default token file does not exist, leaves
-  /// the `token` set to `None`.
+  /// It is not normally necessary to call this, because `Config::new()`
+  /// will automatically laod and use the appropriate token anyway,
+  /// according to the same algorithm.
+  ///
+  /// Arranges that `self.token` is either `Anonymous` or `Value`,
+  /// by establishing a suitable default, and loading the token from
+  /// a file, as necessary.
   #[throws(anyhow::Error)]
   pub fn load_default_token(&mut self) -> &mut Self {
-    if self.token.is_none() {
-      let path = self.default_token_path()?;
-      self.token = Token::load(&path)?;
-    }
+    (||{
+      self.token = Some(match self.get_token_or_default()? {
+        None => TokenConfig::Anonymous,
+        Some(v) => TokenConfig::Value(v),
+      });
+      Ok::<_,AE>(())
+    })().context("load default token")?;
     self
   }
+
+  #[throws(FE)]
+  pub(crate) fn get_token(&self) -> Option<Token> {
+    self.get_token_or_default()
+      .map_err(FE::ClientCreationFailed)?
+  }
 }
 
 /// Main entrypoints once a `Forge` has been constructed.
@@ -415,6 +484,26 @@ pub trait ForgeMethods {
   /// It is not necessary for routine objects such as merge requests,
   /// issues, discussion commonets, etc.
   fn clear_id_caches(&mut self);
+
+  /// Get the repo name string from a URL
+  ///
+  /// For example, `https://salsa.debian.org/iwj/otter` => `iwj/otter`.
+  ///
+  /// Correct results are not guaranteed if the url is not a valid url
+  /// at this forge.  The host might or might not be checked against
+  /// the configured/ host.
+  ///
+  /// Not all git URL formats are supported, but the usual ones are.
+  #[throws(FE)]
+  fn remote_url_to_repo(&self, url: &str) -> String {
+    parse_git_remote_url(url, Some(self.host()))?.path.to_owned()
+  }
+
+  /// Inspect forge's hostname
+  fn host(&self) -> &str;
+
+  /// Inspect forge's hostname
+  fn kind(&self) -> Kind;
 }
 
 /// Main trait; forge objects are `Box<dyn Forge>`
